  import com.hamoid.*;
  
  final float ScreenWidth  = 256;
  final float ScreenHeight = 256;

  final float FrameWidth      = 50;     // width of silent region around QR code
  final float ProgressPerTick = 5.0;                            // drawing speed
  final int   PauseDuration   = 60*4;                         // 4 seconds break

  float [][][] LayerList;             // 2 layers with lines built from vertices

  int   LayerIndex;                     // which layer is currently being drawn?
  int   PauseCounter = 0;                                    // > 0 during pause
  float minX,maxX, minY,maxY;                                    // layer bounds
  int   LineIndex, VertexIndex;     // which line/vertex of line is being drawn?
  float Offset;       // how much of the current segment has already been drawn?
  float X,Y;                                            // current tool position

  PGraphics Background;                                  // shows previous layer
  PGraphics Foreground;                                    // shows active layer

  int         LoopCounter; // save video for loops 1 and 2, not for 0 and others
  VideoExport Video;

  void setup () {
    size(256,256);                                   // ScreenWidth,ScreenHeight
    frameRate(60);
    
    LoopCounter = 0;
    Video = new VideoExport(this, "QR-Code-PrintSimulator.mp4");
    Video.setFrameRate(60);                                     // same as above
    Video.startMovie();

    Background = createGraphics(round(ScreenWidth),round(ScreenHeight));
     Background.beginDraw();
      Background.background(255);
      Background.stroke(200);
      Background.strokeWeight(1.0);
      Background.strokeCap(ROUND);
      Background.noFill();
     Background.endDraw();
    Foreground = createGraphics(round(ScreenWidth),round(ScreenHeight));
     Foreground.beginDraw();
      Foreground.background(255,255,255,0);
      Foreground.stroke(0);
      Foreground.strokeWeight(1.0);
      Foreground.strokeCap(ROUND);
      Foreground.noFill();
     Foreground.endDraw();

    LayerList = loadedLayerList();

    minX = maxX = minY = maxY = 0;
    determineBoundsOf(LayerList[0]);
    determineBoundsOf(LayerList[1]);

    LayerIndex = 1; proceedWithNextLayer();
    loop();
  }

/**** loadedLayerList ****/

  float [][][] loadedLayerList () {
    JSONArray LayerArray = loadJSONArray("encodedLayers.json");

    float [][][] LayerList = new float [2][][];
      LayerList[0] = LineListFrom(LayerArray.getJSONArray(0));
      LayerList[1] = LineListFrom(LayerArray.getJSONArray(1));
    return LayerList;
  }

/**** LineListFrom ****/

  float [][] LineListFrom (JSONArray LineArray) {
    float [][] LineList = new float [LineArray.size()][];
      for (int i = 0; i < LineArray.size(); i++) {
        LineList[i] = VertexListFrom(LineArray.getJSONArray(i));
      };
    return LineList;
  }

/**** VertexListFrom ****/

  float [] VertexListFrom (JSONArray VertexArray) {
    float [] VertexList = new float [VertexArray.size()];
      for (int i = 0; i < VertexArray.size(); i++) {
        VertexList[i] = VertexArray.getFloat(i);
      };
    return VertexList;
  }

/**** determineBoundsOf ****/

  void determineBoundsOf (float [][] LineList) {
    for (int LineIndex = 0; LineIndex < LineList.length; LineIndex++) {
      float[] VertexList = LineList[LineIndex];
      for (int i = 0; i < VertexList.length; i += 2) {
        float X = VertexList[i];
        if (X < minX) { minX = X; }
        if (X > maxX) { maxX = X; }

        float Y = VertexList[i+1];
        if (Y < minY) { minY = Y; }
        if (Y > maxY) { maxY = Y; }
      }
    }
  }

/**** mappedX/Y ****/

  float mappedX (float X) {
    return FrameWidth + (X-minX)/(maxX-minX) * (ScreenWidth-2*FrameWidth);
  }

  float mappedY (float Y) {
    return ScreenHeight - (
      FrameWidth + (Y-minY)/(maxY-minY) * (ScreenHeight-2*FrameWidth)
    );
  }

/**** proceedWithCurrentSegment ****/

  void proceedWithCurrentSegment (float stillToDraw) {
    float[][] LineList   = LayerList[LayerIndex];
    float[]   VertexList = LineList[LineIndex];

    float firstX = VertexList[VertexIndex];
    float firstY = VertexList[VertexIndex+1];
    float lastX  = VertexList[VertexIndex+2];
    float lastY  = VertexList[VertexIndex+3];

    float SegmentLength   = sqrt(sq(lastX-firstX) + sq(lastY-firstY));
    float remainingLength = SegmentLength-Offset;
    if (stillToDraw < remainingLength) {
      float X1 = X; float X2 = X + (lastX-firstX)*stillToDraw/SegmentLength;
      float Y1 = Y; float Y2 = Y + (lastY-firstY)*stillToDraw/SegmentLength;
        drawSegment(X1,Y1, X2,Y2);
      X = X2; Y = Y2; Offset += stillToDraw;
    } else {
      drawSegment(X,Y, lastX,lastY);
      X = lastX; Y = lastY; Offset = 0;

      proceedWithNextSegment(stillToDraw-remainingLength);
    }
  }

/**** proceedWithNextSegment ****/

  void proceedWithNextSegment (float stillToDraw) {
    float[][] LineList   = LayerList[LayerIndex];
    float[]   VertexList = LineList[LineIndex];

    VertexIndex += 2;
    if (VertexIndex + 2 < VertexList.length) {
      X = VertexList[VertexIndex]; Y = VertexList[VertexIndex+1];
      proceedWithCurrentSegment(stillToDraw);
    } else {
      proceedWithNextLine(stillToDraw);
    }
  }

/**** proceedWithNextLine ****/

  void proceedWithNextLine (float stillToDraw) {
    float[][] LineList = LayerList[LayerIndex];

    LineIndex += 1;
    if (LineIndex < LineList.length) {
      float[] VertexList = LineList[LineIndex];

      VertexIndex = 0; Offset = 0;
      X = VertexList[VertexIndex]; Y = VertexList[VertexIndex+1];    // "moveto"
      proceedWithCurrentSegment(stillToDraw);
    } else {
      PauseCounter = 1;                                    // pause for a moment
    }
  }

/**** proceedWithNextLayer ****/

  void proceedWithNextLayer () {
    LayerIndex = (LayerIndex == 0 ? 1 : 0);
    LineIndex = 0; VertexIndex = 0; Offset = 0;
    X = LayerList[LayerIndex][LineIndex][0];
    Y = LayerList[LayerIndex][LineIndex][1];

    Background.background(255);
    Foreground.background(255,255,255,0);
  }

/**** drawSegment ****/

  void drawSegment (float X1,float Y1, float X2,float Y2) {
    Foreground.beginDraw();
      Foreground.line(mappedX(X1),mappedY(Y1), mappedX(X2),mappedY(Y2));
    Foreground.endDraw();

    Background.beginDraw();
      Background.line(mappedX(X1),mappedY(Y1), mappedX(X2),mappedY(Y2));
    Background.endDraw();
  }

/**** drawCrosshair ****/

  void drawCrosshair () {
    stroke(0);
    strokeWeight(1);
    noFill();
    line(0,mappedY(Y), ScreenWidth,mappedY(Y));
    line(mappedX(X),0, mappedX(X),ScreenHeight);
  }

/**** draw ****/

  void draw () {
    if (PauseCounter == 0) {
      proceedWithCurrentSegment(ProgressPerTick);
  
      background(255);
      image(Background, 0,0);
      image(Foreground, 0,0);
  
      drawCrosshair();
      
      if ((LoopCounter == 1) || (LoopCounter == 2)) {
        Video.saveFrame();
      }
    } else {
      PauseCounter += 1;
      if (PauseCounter == 2) {
        background(255);
        image(Background, 0,0);
        image(Foreground, 0,0);
      }
      
      if ((LoopCounter == 1) || (LoopCounter == 2)) {
        Video.saveFrame();
      }
      
      if (PauseCounter > PauseDuration) {
        PauseCounter = 0;
        proceedWithNextLayer();
        
        if (LoopCounter <= 3) {
          LoopCounter++;
          if (LoopCounter == 3) {
            Video.endMovie();
            
            saveFrame("QR-Code-PrintSimulator.png");
            
            LoopCounter++;
            println("Video rendering complete");
          }
        }
      }
    }    
  }
